# system modules
import os, sys, shutil, urllib2

# local modules
import config, packagelist, tarfile

__version__ = "$Revision: 1.18 $"

class Installer:
    """Handles installation of packages"""
    def __init__(self, packageName):
        self.__config = config.getConfig()
        self.__config.resetServerAddress()
        self.__packages = packagelist.getPackageList()
        self.__installed = config.getInstalled()
        self.__package = packageName
        self.__version = self.__packages.getVersion(packageName)
        self.__fileName = self.__package + '-' + self.__version + '.tar.gz'
        self.__size = self.__packages.getSize(packageName)

        self.__buildPath = os.tmpnam()
        os.mkdir(self.__buildPath)
        self.__sourceTree = None
        

    def install(self):
        if (self.__installed.installed(self.__package) and
            self.__installed.current(self.__package, self.__version)):
            sys.stdout.write('%s is current\n' % self.__package)
            return
        cwd = os.getcwd()
        os.chdir(os.path.expanduser('~/.ciphon/'))

        unpackedDir = ""
        getResult = self.__get()

        try:
            unpackedDir = self.__unpack()
        except Exception, details:
            import traceback
            traceback.print_exc()
            try:
                self.__clean(unpackedDir)
            except Exception, details:
                import traceback
                traceback.print_exc()
            raise InstallError('Unpacking package')

        try:
            self.__build()
        except Exception, details:
            import traceback
            traceback.print_exc()
            try:
                self.__clean(unpackedDir)
            except Exception, details:
                import traceback
                traceback.print_exc()
            raise PackageBuildError('Building package')

        try:
            self.__install(unpackedDir)
        except Exception, details:
            import traceback
            traceback.print_exc()
            try:
                self.__clean(unpackedDir)
            except Exception, details:
                import traceback
                traceback.print_exc()
            raise PackageInstallError('Building package')

        try:
            self.__installed.update(self.__package, self.__version)
            self.__installed.write()
            self.__clean(unpackedDir)
        except Exception, details:
            import traceback
            traceback.print_exc()
            raise PackageInstallError('Cleaning package')
        os.chdir(cwd)

    def __get(self):
        """Try to download package"""
        packageLocation = self.__packages.getLocation(self.__package)
        try:
            while 1:
                self.__count = 0
                try:
                    if sys.platform == 'win32':
                        self.__fileHandle = open(self.__fileName, 'wb')
                    else:
                        self.__fileHandle = open(self.__fileName, 'w')
                except OSError:
                    PackageGetError("Can't open file for writing")
                self.__sizeDownloaded = 0
                server = self.__config.getServerAddress()
                url = "%s/%s/%s/%s" % (server['address'],
                                       server['path'],
                                       packageLocation,
                                       self.__fileName)

                sys.stdout.write("Downloading file..\n")
                try:
                    data = urllib2.urlopen(url)
                except urllib2.URLError:
                    sys.stdout.write("Can't get package from %s...\n" % url)
                while 1:
                    buffer = data.read(8192)
                    if buffer == "":
                        break
                    self.__fileHandle.write(buffer)
                break
        except config.ServerError:
            raise PackageGetError("Can't download package from any server!")

        self.__fileHandle.close()
        self.__config.resetServerAddress()

        self.__pkgFile = os.path.join(os.getcwd(), self.__fileName)

        return 0

    def __write(self, block):
        """Write binary blocks to file"""
        self.__sizeDownloaded += len(block)
        sys.stdout.write("*")
        self.__count += 1
        if self.__count == 70:
            percentage = (self.__sizeDownloaded * 100 ) / self.__size
            sys.stdout.write("[%s\%%]\n" % percentage)
        self.__fileHandle.write(block)
        
    def __unpack(self):
        """Unpack a package so that it can be built and installed"""
        
        presentDir = os.getcwd()
        os.chdir(self.__buildPath)
        tarObject = tarfile.TarFile(self.__pkgFile)
        self.__sourceTree = os.path.join(self.__buildPath, tarObject.namelist()[0])
        try:
            tarObject.untar(self.__buildPath)
        except IOError:
            import traceback
            traceback.print_exc()
            os.chdir(presentDir)
            sys.stderr.write("Can't untar package!\n")
            raise PackageBuildError()
        os.chdir(presentDir)
        return self.__sourceTree
    
    def __build(self):
        """Build a package using distutils"""

        presentDir = os.getcwd()
        sys.stdout.write("Changing to downloaed package directory\n")
        os.chdir(self.__sourceTree)

        distutilsExec = "%s setup.py" % sys.executable
        if self.__config.getParameter('installtype') == 'rpm':
            sys.stdout.write("Building rpm package\n")
            returnVal = os.system('%s bdist_rpm' % distutilsExec)
        elif self.__config.getParameter('installtype') == 'win32':
            sys.stdout.write("Building windows package\n")
            returnVal = os.system('%s  bdist_wininst' % distutilsExec)            
        else:
            sys.stdout.write("Building standard package\n")            
            returnVal = os.system('%s  build' % distutilsExec)

        if returnVal == 0:
            os.chdir(presentDir)
            return
        else:
            os.chdir(presentDir)
            sys.stderr.write("An error occured while running setup\n")
            raise PackageBuildError()

    def __install(self, unpackedDir):
        """Install a package using disutils"""
        presentDir = os.getcwd()
        os.chdir(os.path.join(unpackedDir, 'dist'))

        if self.__config.getParameter('installtype') == 'rpm':
            sys.stdout.write("Trying to rpm install package\n")
            returnVal = os.system('rpm -Uvh %s-%s-1.noarch.rpm' % (self.__package,
                                                                   self.__version))
        elif self.__config.getParameter('installtype') == 'win32':
            installer = '%s-%s.win32-py%s.%s.exe' % \
                        (self.__package, self.__version,
                         sys.version_info[0], sys.version_info[1])
            returnVal = os.system(installer)
            # XXX returnVal is meaningless for current bdist_wininst installers.
        else:
            sys.stdout.write("Installing package\n")
            returnVal = os.system('%s setup.py install' % sys.executable)

        if returnVal == 0:
            os.chdir(presentDir)
            return
        else:
            os.chdir(presentDir)
            sys.stderr.write("An error occured while trying to install module\n")
            raise PackageInstallError()

    def __clean(self, unpackedDir):
        """Cleans up a package after installation"""
        try:
            os.unlink(self.__fileName)
            shutil.rmtree(self.__buildPath)
        except Exception:
            sys.stderr.write("An error occured while trying to clean up\n")
            sys.stderr.write("The package has been installed but you may\n")
            sys.stderr.write("want to clean up the build directory\n")            
        
def main(input):
    """Handle package install requests"""
    try:
        installer = Installer(input[1])
        installer.install()
    except InstallError, x:
        sys.stderr.write('Error installing package: %s\nGiving up...\n' % x);
    except packagelist.NotFound:
        sys.stderr.write('No such package exists!\n');



def installModule(commands, help):
    """Install the install command into the main eval loop"""

    commands['install'] = main
    help['install'] = "install [package]"

# ************* Exceptions thrown by Installer class *****************

class InstallError(Exception):
    """Superclass for all exceptions relating to installation errors"""
    pass

class PackageInstallError(InstallError):
    """Error installing a package"""
    pass

class PackageBuildError(InstallError):
    """ Error building a package"""
    pass

class PackageGetError(InstallError):
    """Error downloading or otherwise obtaining a package"""
    pass
